#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "traverse.h"
#include "ftable.h"

bool handler(void* handle, char* path, struct stat* statbuf) {
   FileTable* ft = (FileTable*) handle;
   ftable_add(ft, statbuf->st_dev, statbuf->st_ino, path);
   if (S_ISLNK(statbuf->st_mode)) {
      struct stat rbuf;
      if (stat(path, &rbuf) == 0) {
	 ftable_add_symlink(ft, rbuf.st_dev, rbuf.st_ino, path);
      }
   }
   return true;
}

void print_links(NameList* list, unsigned int indent) {
   while (list) {
      printf("%*s%s\n", indent, "", list->path);
      list = list->next;
   }
}

int main(int argc, char** argv) {
   char* cmdname = *argv;
   char usage[] = "Usage: %s [-h] [-s] [dirpath]\n";
   bool show_symlinks = false;
   bool show_hardlinks = false;

   int option;
   while ((option = getopt(argc, argv, "hs")) != -1) {
      switch (option) {
	 case 'h':
	    show_hardlinks = true; break;
	 case 's':
	    show_symlinks = true; break;
	 default:
	    fprintf(stderr, usage, cmdname); exit(1);
      }
   }
   argv += optind; argc -= optind;
   if (argc > 1) {
      fprintf(stderr, usage, cmdname); exit(1);
   }
   char* path = ".";
   if (argc > 0) path = argv[0];

   bool show_both = (show_symlinks == show_hardlinks);
   if (show_both) {
      show_symlinks = show_hardlinks = true;
   }
   unsigned int indent = show_both? 6: 3;

   FileTable* ft = ftable_create();
   if (!ft) {
      fprintf(stderr, "%s: out of memory\n", cmdname); exit(1);
   }

   /* run recursive traverse */
   traverse(path, ft, handler);

   /* iterate through all entries of our file table
      and print those that have links
   */
   ftable_start(ft);
   FileTableEntry* entry;
   while ((entry = ftable_next(ft))) {
      /* skip uninteresting cases */
      if (!entry->hardlinks) continue;
      if (!entry->hardlinks->next && !entry->symlinks) continue;
      if (!show_hardlinks && !entry->symlinks) continue;
      if (!show_symlinks && !entry->hardlinks->next) continue;

      /* print one of the hardlinked names */
      puts(entry->hardlinks->path);

      /* print all other links of interest to it */
      NameList* name = entry->hardlinks->next;
      if (show_hardlinks && name) {
	 if (show_both) printf("   hard links:\n");
	 print_links(name, indent);
      }
      name = entry->symlinks;
      if (show_symlinks && name) {
	 if (show_both) printf("   symbolic links:\n");
	 print_links(name, indent);
      }
   }
}
