#include <stdlib.h>
#include <string.h>
#include "ftable.h"

#define BUCKET_LEN 4

struct FileTable {
   FileTableEntry* bucket[BUCKET_LEN];
   unsigned int it_index; FileTableEntry* it_next;
};

static unsigned int hash(dev_t device, ino_t inode) {
   return (device ^ inode) % BUCKET_LEN;
}

FileTable* ftable_create() {
   FileTable* ft = malloc(sizeof(FileTable));
   if (ft) {
      for (unsigned int index = 0; index < BUCKET_LEN; ++index) {
	 ft->bucket[index] = 0;
      }
      ft->it_index = 0; ft->it_next = 0;
   }
   return ft;
}

static FileTableEntry* find(FileTable* ft, dev_t device, ino_t inode) {
   unsigned int index = hash(device, inode);
   FileTableEntry* entry = ft->bucket[index];
   while (entry && (entry->device != device || entry->inode != inode)) {
      entry = entry->next;
   }
   if (!entry) {
      entry = malloc(sizeof(FileTableEntry));
      if (entry) {
	 entry->device = device; entry->inode = inode;
	 entry->hardlinks = entry->symlinks = 0;
	 entry->next = ft->bucket[index];
	 ft->bucket[index] = entry;
      }
   }
   return entry;
}

static bool add_name(NameList** list, char* path) {
   NameList* member = malloc(sizeof(NameList));
   if (!member) return false;
   member->path = strdup(path);
   member->next = *list;
   *list = member;
   return true;
}

bool ftable_add(FileTable* ft, dev_t device, ino_t inode, char* path) {
   FileTableEntry* entry = find(ft, device, inode);
   if (!entry) return false;
   return add_name(&entry->hardlinks, path);
}

bool ftable_add_symlink(FileTable* ft, dev_t device, ino_t inode,
      char* symlink_path) {
   FileTableEntry* entry = find(ft, device, inode);
   if (!entry) return false;
   return add_name(&entry->symlinks, symlink_path);
}

void ftable_start(FileTable* ft) {
   ft->it_index = 0; ft->it_next = 0;
}

FileTableEntry* ftable_next(FileTable* ft) {
   while (ft->it_next == 0 && ft->it_index < BUCKET_LEN) {
      ft->it_next = ft->bucket[ft->it_index++];
   }
   FileTableEntry* entry = ft->it_next;
   if (entry) ft->it_next = entry->next;
   return entry;
}
